
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>

#include "struct.h"
#include "io.h"
#include "config.h"
#include "server.h"

#define BUFSIZE 2048
#define CONNECT "Connected to MTPserver " VERSION
#define WIZ_LEVEL 21

char *filebuffer;
struct mtpheader  *header;
struct mtpmessage *mtpm;

char mtphost[BUFSIZE];
int mtpport = MTPPORT;

/*------------------------------------------------------------------------*/

int main(int argc, char **argv)
{
   int fh, sh, ok, cycle, is_write;
   char *path, tmp[300], tmp2[100];
   unsigned short filelength;

   for (cycle = 0; cycle < MAX_CYCLE; cycle++)
   {
      if (!fork())              /* Childprozess? */
      {
	 gethostname(mtphost, BUFSIZE);
	 if (argc == 2)         /* MTPPort umsetzen? */
	    mtpport = atoi(argv[1]);

	 /* MTP Header, Filebuffer, und Messagebuffer allozieren */
	 open_all();
	 /* Log Start of MTP Server in MTPLOGFILE */ 
	 init_server();

	 sh = 0;  

	 for (;;)               /* Endlosschleife fuer jedes File */
	 {
	    if (sh)             /* Alten Socket schliessen */
	       close(sh);
	                        /* Neuer Socket, auf Connection warten */
	    if (!(sh = create_socket(mtpport)))
	    {
	       close_all();
	       exit(0);
	    }
	                        /* Testzeichen empfangen */
	    safe_read(sh, tmp, 1);
	    if (*tmp != '@')
	       continue;

	                        /* Header empfangen */
	    if (!(safe_read(sh, (char *)header, sizeof(struct mtpheader))))
	         continue;

	                        /* Level des Wizards pruefen */
	    if (grant_access(header) < WIZ_LEVEL)
	    {
	       sprintf(tmp,
		       CONNECT " at %s.\nPermission denied for user '%s'.\n",
		       mtphost, header->player);
	       write_message(sh, mtpm, tmp, 1);
	       continue;
	    }
	                        /* Absoluter Pfad auf Host */
	    path = build_path(header);
	                        /* Uebertragungsmodus */
	    is_write = !(chars2short(header->flags) & RECEIVE);
	                        /* Zugriff auf File pruefen */
	    ok = valid_access(path, header->filename, header->player, is_write);
	                        /* Filelaenge aus Header lesen */
	    filelength = chars2short(header->filelength);

	    sprintf(tmp, CONNECT " at %s.\n", mtphost);

	    if (ok)             /* Message fuer Client vorbereiten */
	    {
	       sprintf(tmp2,
		       (is_write ? "Upload of '%s' to mud ..." :
		                   "Download of '%s' from mud ..."),
		       header->filename);
	       strcat(tmp, tmp2);
	    }
	    else
	    {
	       sprintf(tmp2,   /* Kein Zugriff, also Meldung und cu */
	               (is_write ? "Write access to '%s' denied.\n" :
		                   "Read access to '%s' denied.\n"),
		       header->filename);
	       strcat(tmp, tmp2);
	       write_message(sh, mtpm, tmp, 1);
	       continue;
	    }

	    if (check_directory(path))
	    {
	       strcat(tmp, " aborted.\nDestination file is a directory!\n");
	       write_message(sh, mtpm, tmp, 1);
	       continue;
	    }
			        /* SEND FILE TO CLIENT */
	    if (chars2short(header->flags) & RECEIVE)
	    {
	       send_file(sh, path, header, tmp);
	       continue;
	    }
	                       /* RECEIVE FILE FROM CLIENT */
	                       /* Open/Create file on serverhost */
	    fh = mtp_open(path, header);
	    if (fh < 1)
	    {
	       strcat(tmp, " aborted.\nCouldn't write destination file" 
	                   " (maybe already existing ?)\n");
	       write_message(sh, mtpm, tmp, 1);
	       continue;
	    }
	                        /* Send info to client */
	    if (!(write_message(sh, mtpm, tmp, 0)))
	       continue;
	                        /* Receive file from client */
	    if (!(safe_read(sh, filebuffer, filelength)))
	       continue;
	                        /* Write file to disk */
	    safe_write(fh, filebuffer, filelength);
	                        /* Send info to client, and quit client */
	    sprintf(tmp, " %d Bytes completed.\n", filelength);
	    write_message(sh, mtpm, tmp, 1);
	                        /* Log filetransmission to MTPLOGFILE */
	    log_message(header);

	    if (fh)             /* Close local file */
	       close(fh);
	 }			/* main loop */
      }
      wait(0);
      sleep(2);
   }
   exit(1);
}

/*-------------------------------------------------------------------------*/

void open_all(void)
{
   header = (struct mtpheader *) introduce(sizeof(struct mtpheader));

   filebuffer = introduce(MAX_FILELENGTH + 1);
   mtpm = (struct mtpmessage *) introduce(sizeof(struct mtpmessage));
}

/*-------------------------------------------------------------------------*/

void close_all(void)
{
   dismiss((char*)header);
   dismiss(filebuffer);
   dismiss((char *)mtpm);
}

/*-------------------------------------------------------------------------*/

void init_server(void)
{
   strcpy(header->player, "MTPmaster");
   strcpy(header->filename, "MTPserver initialized");
   short2chars(0, header->filelength);
   short2chars(0, header->flags);
   log_message(header);
}

/*-------------------------------------------------------------------------*/

int grant_access(struct mtpheader *mtph)
{
   FILE *fd;
   char passwd[20], tmp[30], tmp2[30];
   int i, ok, level;

   ok = 0;
   level = 0;
   sprintf(tmp, "%s/var/players/%c/%s.o", MUDPATH, mtph->player[0], mtph->player);
   if (!(fd = fopen(tmp, "r")))
      return 0;
   while (fscanf(fd, "%29s", tmp) != EOF)
   {
      if (!strcmp(tmp, "level"))
	 fscanf(fd, "%d", &level);

      if (!strcmp(tmp, "password"))
      {
	 fscanf(fd, "%29s", tmp2);
	 for (i = 0; i < strlen(tmp2) - 2; i++)
	    passwd[i] = tmp2[i + 1];
	 passwd[i] = '\0';
	 ok = !strcmp(passwd, (char *) crypt(mtph->passwd, passwd));
      }
   }
   if (fd)
      fclose(fd);
   return ok ? level : 0;
}
